#include <cmath>
#include <iostream>
#include <sstream>
#include <sys/time.h>

#define GRN "\33[92m"
#define YEL "\33[93m"
#define RES "\33[0m"

template<typename T>
std::string toStr(const T& nr)
{
  std::ostringstream os;
  os << nr;
  return os.str();
}

template<typename T>
T strTo(const std::string &s)
{
  T num;
  std::istringstream is(s);
  is >> num;
  return num;
}

template<typename T>
T strTo(const char* s)
{
  T num;
  std::istringstream is(s);
  is >> num;
  return num;
}

class TimeCPU
{
  private:
    struct timeval ts;
    double t;

  public:
    TimeCPU()
    : t(0.0)
    {}

    void start()
    {
      gettimeofday(&ts, NULL);
    }

    double read()
    {
      struct timeval te;
      gettimeofday(&te, NULL);
      t = (double)(te.tv_sec-ts.tv_sec)+0.000001*(double)(te.tv_usec-ts.tv_usec);
      return t;
    }

    const double& time() const { return t; }
    double time() { return t; }
};

inline std::ostream& operator<<(std::ostream& os, const TimeCPU& t)
{
  return os << t.time();
}

inline double operator/(const double& l, const TimeCPU& r)
{
  return l/r.time();
}

inline double operator/(const TimeCPU& l, const double& r)
{
  return l.time()/r;
}

inline double operator*(const double& l, const TimeCPU& r)
{
  return l*r.time();
}

inline double operator*(const TimeCPU& l, const double& r)
{
  return l.time()*r;
}

inline double operator+(const double& l, const TimeCPU& r)
{
  return l+r.time();
}

inline double operator+(const TimeCPU& l, const double& r)
{
  return l.time()+r;
}

inline double operator-(const double& l, const TimeCPU& r)
{
  return l-r.time();
}

inline double operator-(const TimeCPU& l, const double& r)
{
  return l.time()-r;
}

double fun(const int N, const int M)
{
  double test=0.0;
  for(int i=0; i<N; ++i)
  {
    for(int j=0; j<M; ++j)
    {
      test = ((double)i*(double)j+test)/8192.0;
    }
  }
  return test;
}

int main()
{
  std::cout << GRN << "-----=====#####=====-----\n" << RES;
  TimeCPU tCPU;
  const int N=65536;
  const int M=65536;
  double outcome;

  tCPU.start();
  outcome = fun(N, M);
  tCPU.read();

  const double Gflop=(double)N*(double)M*3*0.000000001;
  const double perf=Gflop/tCPU;

  std::cout << "Result " << outcome;
  if( fabs(outcome-524336.0)<0.5 )
    std::cout << GRN << " - correct\n" << RES;
  else
    std::cout << YEL << " - error\n" << RES;
  std::cout << "The fun() function executed in " << tCPU << " [s]\n";
  std::cout << "Executed " << Gflop << " billion operations\n";
  std::cout << "Efficiency " << YEL << perf << RES << " [Gflop/s]\n";
  std::cout << GRN << "-----=====#####=====-----\n" << RES;
  return 0;
}
